@@ -1,6 +1,4 @@
-if RUBY_VERSION < '1.9'
-  require 'rubygems'
-end
+RUBY_ENGINE = 'unknown' unless defined? RUBY_ENGINE
 require 'strscan'
 require 'set'
 
@@ -12,20 +10,18 @@ $:.unshift(File.dirname(__FILE__))
 # Asciidoc documents comprise a header followed by zero or more sections.
 # Sections are composed of blocks of content.  For example:
 #
-#   Doc Title
-#   =========
+#   = Doc Title
 #
-#   SECTION 1
-#   ---------
+#   == Section 1
 #
 #   This is a paragraph block in the first section.
 #
-#   SECTION 2
+#   == Section 2
 #
 #   This section has a paragraph block and an olist block.
 #
-#   1. Item 1
-#   2. Item 2
+#   . Item 1
+#   . Item 2
 #
 # Examples:
 #
@@ -89,6 +85,27 @@ module Asciidoctor
 
   end
 
+  # Flags to control compliance with the behavior of AsciiDoc
+  module Compliance
+    # AsciiDoc supports both single-line and underlined
+    # section titles.
+    # This option disables the underlined variant.
+    # Compliance value: true
+    @underline_style_section_titles = true
+    class << self
+      attr_accessor :underline_style_section_titles
+    end
+
+    # Asciidoctor will recognize commonly-used Markdown syntax
+    # to the degree it does not interfere with existing
+    # AsciiDoc syntax and behavior.
+    # Compliance value: false
+    @markdown_syntax = true
+    class << self
+      attr_accessor :markdown_syntax
+    end
+  end
+
   # The root path of the Asciidoctor gem
   ROOT_PATH = File.expand_path(File.join(File.dirname(__FILE__), '..'))
 
@@ -98,6 +115,12 @@ module Asciidoctor
   # Ruby 1.8 doesn't seem to experience this problem (perhaps because it isn't validating the encodings)
   FORCE_ENCODING = RUBY_VERSION > '1.9' && Encoding.default_external != Encoding::UTF_8
 
+  # Flag to indicate that line length should be calculated using a unicode mode hint
+  FORCE_UNICODE_LINE_LENGTH = RUBY_VERSION < '1.9'
+
+  # The endline character to use when rendering output
+  EOL = "\n"
+
   # The default document type
   # Can influence markup generated by render templates
   DEFAULT_DOCTYPE = 'article'
@@ -128,6 +151,15 @@ module Asciidoctor
     'markdown' => '.md'
   }
 
+  # Set of file extensions recognized as AsciiDoc documents (stored as a truth hash)
+  ASCIIDOC_EXTENSIONS = {
+    '.asciidoc' => true,
+    '.adoc' => true,
+    '.ad' => true,
+    '.asc' => true,
+    '.txt' => true
+  }
+
   SECTION_LEVELS = {
     '=' => 0,
     '-' => 1,
@@ -160,23 +192,30 @@ module Asciidoctor
     '~~~'  => [:fenced_code, Set.new]
   }
 
+  DELIMITED_BLOCK_LEADERS = DELIMITED_BLOCKS.keys.map {|key| key[0..1] }.to_set
+
   BREAK_LINES = {
-    %q{'''} => :ruler,
-    '<<<'   => :page_break
+    '\'' => :ruler,
+    '-'  => :ruler,
+    '*'  => :ruler,
+    '_'  => :ruler,
+    '<'  => :page_break
   }
 
-  LIST_CONTEXTS = [:ulist, :olist, :dlist, :colist]
+  #LIST_CONTEXTS = [:ulist, :olist, :dlist, :colist]
 
   NESTABLE_LIST_CONTEXTS = [:ulist, :olist, :dlist]
 
-  ORDERED_LIST_STYLES = [:arabic, :loweralpha, :lowerroman, :upperalpha, :upperroman]
+  # TODO validate use of explicit style name above ordered list (this list is for selecting an implicit style)
+  ORDERED_LIST_STYLES = [:arabic, :loweralpha, :lowerroman, :upperalpha, :upperroman] #, :lowergreek]
 
   ORDERED_LIST_MARKER_PATTERNS = {
     :arabic => /\d+[.>]/,
     :loweralpha => /[a-z]\./,
-    :upperalpha => /[A-Z]\./,
     :lowerroman => /[ivx]+\)/,
+    :upperalpha => /[A-Z]\./,
     :upperroman => /[IVX]+\)/
+    #:lowergreek => /[a-z]\]/
   }
 
   ORDERED_LIST_KEYWORDS = {
@@ -184,12 +223,19 @@ module Asciidoctor
     'lowerroman' => 'i',
     'upperalpha' => 'A',
     'upperroman' => 'I'
+    #'lowergreek' => 'a'
+    #'arabic'     => '1'
+    #'decimal'    => '1'
   }
 
   LIST_CONTINUATION = '+'
 
   LINE_BREAK = ' +'
 
+  # attributes which be changed within the content of the document (but not
+  # header) because it has semantic meaning; ex. numbered
+  FLEXIBLE_ATTRIBUTES = %w(numbered)
+
   # NOTE allows for empty space in line as it could be left by the template engine
   BLANK_LINE_PATTERN = /^[[:blank:]]*\n/
 
@@ -215,10 +261,15 @@ module Asciidoctor
     # Compliance value: false
     :congruent_block_delimiters => true,
 
-    # AsciiDoc will recognize commonly-used Markdown syntax
-    # to the degree it does not interfere with existing
-    # AsciiDoc behavior.
-    :markdown_syntax => true
+    # AsciiDoc drops lines that contain references to missing attributes.
+    # This behavior is not intuitive to most writers
+    # Compliance value: 'drop-line'
+    :attribute_missing => 'skip',
+
+    # AsciiDoc drops lines that contain an attribute unassignemnt.
+    # This behavior may need to be tuned depending on the circumstances.
+    # Compliance value: 'drop-line'
+    :attribute_undefined => 'drop-line',
   }
 
   # The following pattern, which appears frequently, captures the contents between square brackets,
@@ -251,27 +302,18 @@ module Asciidoctor
     # [[:graph:]] is a non-blank character
     :any_list         => /^(?:
                              <?\d+>[[:blank:]]+[[:graph:]]|
-                             [[:blank:]]*(?:(?:-|\*|\.){1,5}|\d+\.|[A-Za-z]\.|[IVXivx]+\))[[:blank:]]+[[:graph:]]|
+                             [[:blank:]]*(?:-|(?:\*|\.){1,5}|\d+\.|[A-Za-z]\.|[IVXivx]+\))[[:blank:]]+[[:graph:]]|
                              [[:blank:]]*.*?(?::{2,4}|;;)(?:[[:blank:]]+[[:graph:]]|$)
                            )/x,
 
     # :foo: bar
     # :Author: Dan
     # :numbered!:
-    :attr_entry       => /^:(\w.*?):(?:[[:blank:]]+(.*))?$/,
-
-    # {name?value}
-    :attr_conditional => /^\s*\{([^\?]+)\?\s*([^\}]+)\s*\}/,
-
-    # +   Attribute values treat lines ending with ' +' as a continuation,
-    #     not a line-break as elsewhere in the document, where this is
-    #     a forced line break. This should be the same regexp as :line_break,
-    #     below, but it gets its own entry because readability ftw, even
-    #     though repeating regexps ftl.
-    :attr_continue    => /^[[:blank:]]*(.*)[[:blank:]]\+[[:blank:]]*$/,
-
-    # :foo!:
-    :attr_delete      => /^:([^:]+)!:$/,
+    # :long-entry: Attribute value lines ending in ' +'
+    #              are joined together as a single value,
+    #              collapsing the line breaks and indentation to
+    #              a single space.
+    :attr_entry       => /^:(!?\w.*?):(?:[[:blank:]]+(.*))?$/,
 
     # An attribute list above a block element
     #
@@ -281,14 +323,16 @@ module Asciidoctor
     # [NOTE, caption="Good to know"]
     # Can be defined by an attribute
     # [{lead}]
-    :blk_attr_list    => /^\[(|[[:blank:]]*[\w\{,.#"'].*)\]$/,
+    :blk_attr_list    => /^\[(|[[:blank:]]*[\w\{,.#"'%].*)\]$/,
 
     # block attribute list or block id (bulk query)
-    :attr_line        => /^\[(|[[:blank:]]*[\w\{,.#"'].*|\[[^\[\]]*\])\]$/,
+    :attr_line        => /^\[(|[[:blank:]]*[\w\{,.#"'%].*|\[[^\[\]]*\])\]$/,
 
     # attribute reference
     # {foo}
     # {counter:pcount:1}
+    # {set:foo:bar}
+    # {set:name!}
     :attr_ref         => /(\\)?\{((set|counter2?):.+?|\w+(?:[\-]\w+)*)(\\)?\}/,
 
     # The author info line the appears immediately following the document title
@@ -299,11 +343,14 @@ module Asciidoctor
     :biblio_macro     => /\\?\[\[\[([\w:][\w:.-]*?)\]\]\]/,
 
     # callout reference inside literal text
-    # <1>
-    # special characters will already be replaced, hence their use in the regex
-    :callout_render   => /\\?&lt;(\d+)&gt;/,
+    # <1> (optionally prefixed by //, # or ;; line comment chars)
+    # <1> <2> (multiple callouts on one line)
+    # <!--1--> (for XML-based languages)
+    # special characters are already be replaced at this point during render
+    :callout_render     => /(?:(?:\/\/|#|;;) ?)?(\\)?&lt;!?(--|)(\d+)\2&gt;(?=(?: ?\\?&lt;!?\2\d+\2&gt;)*$)/,
     # ...but not while scanning
-    :callout_scan     => /\\?<(\d+)>/,
+    :callout_quick_scan => /\\?<!?(--|)(\d+)\1>(?=(?: ?\\?<!?\1\d+\1>)*$)/,
+    :callout_scan       => /(?:(?:\/\/|#|;;) ?)?(\\)?<!?(--|)(\d+)\2>(?=(?: ?\\?<!?\2\d+\2>)*$)/,
 
     # <1> Foo
     :colist           => /^<?(\d+)>[[:blank:]]+(.*)/,
@@ -317,7 +364,7 @@ module Asciidoctor
     :comment          => %r{^//(?:[^/]|$)},
 
     # one,two;three;four
-    :ssv_or_csv_delim   => /,|;/,
+    :ssv_or_csv_delim => /,|;/,
 
     # one two	three
     :space_delim      => /([^\\])[[:blank:]]+/,
@@ -340,21 +387,27 @@ module Asciidoctor
     #   That which precedes 'bar' (see also, <<bar>>)
     # The term may be an attribute reference
     # {term_foo}:: {def_foo}
-    # REVIEW leading space has already been stripped, so may not need in regex
-    :dlist            => /^[[:blank:]]*(.*?)(:{2,4}|;;)(?:[[:blank:]]+(.*))?$/,
+    # NOTE negative match for comment line is intentional since that isn't handled when looking for next list item
+    # QUESTION should we check for line comment in regex or when scanning the lines?
+    :dlist            => /^(?!\/\/)[[:blank:]]*(.*?)(:{2,4}|;;)(?:[[:blank:]]+(.*))?$/,
     :dlist_siblings   => {
                            # (?:.*?[^:])? - a non-capturing group which grabs longest sequence of characters that doesn't end w/ colon
-                           '::' => /^[[:blank:]]*((?:.*[^:])?)(::)(?:[[:blank:]]+(.*))?$/,
-                           ':::' => /^[[:blank:]]*((?:.*[^:])?)(:::)(?:[[:blank:]]+(.*))?$/,
-                           '::::' => /^[[:blank:]]*((?:.*[^:])?)(::::)(?:[[:blank:]]+(.*))?$/,
-                           ';;' => /^[[:blank:]]*(.*)(;;)(?:[[:blank:]]+(.*))?$/
+                           '::' => /^(?!\/\/)[[:blank:]]*((?:.*[^:])?)(::)(?:[[:blank:]]+(.*))?$/,
+                           ':::' => /^(?!\/\/)[[:blank:]]*((?:.*[^:])?)(:::)(?:[[:blank:]]+(.*))?$/,
+                           '::::' => /^(?!\/\/)[[:blank:]]*((?:.*[^:])?)(::::)(?:[[:blank:]]+(.*))?$/,
+                           ';;' => /^(?!\/\/)[[:blank:]]*(.*)(;;)(?:[[:blank:]]+(.*))?$/
                          },
 
+    :illegal_sectid_chars => /&(?:[[:alpha:]]+|#[[:digit:]]+|#x[[:alnum:]]+);|\W+?/,
+
     # footnote:[text]
     # footnoteref:[id,text]
     # footnoteref:[id]
     :footnote_macro   => /\\?(footnote|footnoteref):\[((?:\\\]|[^\]])*?)\]/,
 
+    # gist::123456[]
+    :generic_blk_macro => /^(\w[\w\-]*)::(\S+?)\[((?:\\\]|[^\]])*?)\]$/,
+
     # kbd:[F3]
     # kbd:[Ctrl+Shift+T]
     # kbd:[Ctrl+\]]
@@ -375,8 +428,10 @@ module Asciidoctor
     :media_blk_macro  => /^(image|video|audio)::(\S+?)\[((?:\\\]|[^\]])*?)\]$/,
 
     # image:filename.png[Alt Text]
+    # image:http://example.com/images/filename.png[Alt Text]
     # image:filename.png[More [Alt\] Text] (alt text becomes "More [Alt] Text")
-    :image_macro      => /\\?image:([^:\[]+)\[((?:\\\]|[^\]])*?)\]/,
+    # icon:github[large]
+    :image_macro      => /\\?(?:image|icon):([^:\[][^\[]*)\[((?:\\\]|[^\]])*?)\]/,
 
     # indexterm:[Tigers,Big cats]
     # (((Tigers,Big cats)))
@@ -403,7 +458,7 @@ module Asciidoctor
 
     # inline link and some inline link macro
     # FIXME revisit!
-    :link_inline      => %r{(^|link:|\s|>|&lt;|[\(\)\[\]])(\\?(?:https?|ftp|irc)://[^\s\[<]*[^\s.,\[<])(?:\[((?:\\\]|[^\]])*?)\])?},
+    :link_inline      => %r{(^|link:|\s|>|&lt;|[\(\)\[\]])(\\?(?:https?|ftp|irc)://[^\s\[\]<]*[^\s.,\[\]<])(?:\[((?:\\\]|[^\]])*?)\])?},
 
     # inline link macro
     # link:path[label]
@@ -429,6 +484,12 @@ module Asciidoctor
     # <<< (pagebreak)
     :break_line        => /^('|<){3,}$/,
 
+    # ''' or ' ' ' (ruler)
+    # --- or - - - (ruler)
+    # *** or * * * (ruler)
+    # <<< (pagebreak)
+    :break_line_plus   => /^(?:'|<){3,}$|^ {0,3}([-\*_])( *)\1\2\1$/,
+
     # inline passthrough macros
     # +++text+++
     # $$text$$
@@ -441,7 +502,7 @@ module Asciidoctor
 
     # inline literal passthrough macro
     # `text`
-    :pass_lit         => /(^|[^`\w])(\\?`([^`\s]|[^`\s].*?\S)`)(?![`\w])/m,
+    :pass_lit         => /(^|[^`\w])(?:\[([^\]]+?)\])?(\\?`([^`\s]|[^`\s].*?\S)`)(?![`\w])/m,
 
     # placeholder for extracted passthrough text
     :pass_placeholder => /\e(\d+)\e/,
@@ -460,7 +521,7 @@ module Asciidoctor
     :illegal_attr_name_chars => /[^\w\-]/,
 
     # 1*h,2*,^3e
-    :table_colspec    => /^(?:(\d+)\*)?([<^>](?:\.[<^>]?)?|(?:[<^>]?\.)?[<^>])?(\d+)?([a-z])?$/,
+    :table_colspec    => /^(?:(\d+)\*)?([<^>](?:\.[<^>]?)?|(?:[<^>]?\.)?[<^>])?(\d+%?)?([a-z])?$/,
 
     # 2.3+<.>m
     # TODO might want to use step-wise scan rather than this mega-regexp
@@ -469,6 +530,10 @@ module Asciidoctor
       :end => /[[:blank:]]+(?:(\d+(?:\.\d*)?|(?:\d*\.)?\d+)([*+]))?([<^>](?:\.[<^>]?)?|(?:[<^>]?\.)?[<^>])?([a-z])?$/
     },
 
+    # docbook45
+    # html5
+    :trailing_digit   => /[[:digit:]]+$/,
+
     # .Foo   but not  . Foo or ..Foo
     :blk_title        => /^\.([^\s.].*)$/,
 
@@ -535,9 +600,13 @@ module Asciidoctor
     # http://domain
     # https://domain
     # data:info
-    :uri_sniff        => /^[[:alpha:]][[:alnum:].+-]*:/i,
+    :uri_sniff        => %r{\A[[:alpha:]][[:alnum:].+-]*:/*},
 
-    :uri_encode_chars => /[^\w\-.!~*';:@=+$,()\[\]]/
+    :uri_encode_chars => /[^\w\-.!~*';:@=+$,()\[\]]/,
+
+    :mantitle_manvolnum => /^(.*)\((.*)\)$/,
+
+    :manname_manpurpose => /^(.*?)[[:blank:]]+-[[:blank:]]+(.*)$/
   }
 
   INTRINSICS = Hash.new{|h,k| STDERR.puts "Missing intrinsic: #{k.inspect}"; "{#{k}}"}.merge(
@@ -668,16 +737,15 @@ module Asciidoctor
   # options - a String, Array or Hash of options to control processing (default: {})
   #           String and Array values are converted into a Hash.
   #           See Asciidoctor::Document#initialize for details about options.
-  # block   - a callback block for handling include::[] directives
   #
   # returns the Asciidoctor::Document
-  def self.load(input, options = {}, &block)
+  def self.load(input, options = {})
     if (monitor = options.fetch(:monitor, false))
       start = Time.now
     end
 
     attrs = (options[:attributes] ||= {})
-    if attrs.is_a? Hash
+    if attrs.is_a?(Hash) || (RUBY_ENGINE == 'jruby' && attrs.is_a?(Java::JavaUtil::Map))
       # all good; placed here as optimization
     elsif attrs.is_a? Array
       attrs = options[:attributes] = attrs.inject({}) do |accum, entry|
@@ -695,8 +763,15 @@ module Asciidoctor
         accum[k] = v || ''
         accum
       end
+    elsif attrs.respond_to?('keys') && attrs.respond_to?('[]')
+      # convert it to a Hash as we know it
+      original_attrs = attrs
+      attrs = options[:attributes] = {}
+      original_attrs.keys.each do |key|
+        attrs[key] = original_attrs[key]
+      end
     else
-      raise ArgumentError, 'illegal type for attributes option'
+      raise ArgumentError, "illegal type for attributes option: #{attrs.class.ancestors}"
     end
 
     lines = nil
@@ -727,7 +802,7 @@ module Asciidoctor
       start = Time.now
     end
 
-    doc = Document.new(lines, options, &block) 
+    doc = Document.new(lines, options) 
     if monitor
       parse_time = Time.now - start
       monitor[:read] = read_time
@@ -747,11 +822,10 @@ module Asciidoctor
   # options - a String, Array or Hash of options to control processing (default: {})
   #           String and Array values are converted into a Hash.
   #           See Asciidoctor::Document#initialize for details about options.
-  # block   - a callback block for handling include::[] directives
   #
   # returns the Asciidoctor::Document
-  def self.load_file(filename, options = {}, &block)
-    Asciidoctor.load(File.new(filename), options, &block)
+  def self.load_file(filename, options = {})
+    Asciidoctor.load(File.new(filename), options)
   end
 
   # Public: Parse the AsciiDoc source input into an Asciidoctor::Document and render it
@@ -780,11 +854,10 @@ module Asciidoctor
   # options - a String, Array or Hash of options to control processing (default: {})
   #           String and Array values are converted into a Hash.
   #           See Asciidoctor::Document#initialize for details about options.
-  # block   - a callback block for handling include::[] directives
   #
   # returns the Document object if the rendered result String is written to a
   # file, otherwise the rendered result String
-  def self.render(input, options = {}, &block)
+  def self.render(input, options = {})
     in_place = options.delete(:in_place) || false
     to_file = options.delete(:to_file)
     to_dir = options.delete(:to_dir)
@@ -803,7 +876,7 @@ module Asciidoctor
       options[:header_footer] = true
     end
 
-    doc = Asciidoctor.load(input, options, &block)
+    doc = Asciidoctor.load(input, options)
 
     if to_file == '/dev/null'
       return doc
@@ -852,7 +925,7 @@ module Asciidoctor
       if stream_output
         to_file.write output.rstrip
         # ensure there's a trailing endline
-        to_file.write "\n"
+        to_file.write EOL
       else
         File.open(to_file, 'w') {|file| file.write output }
         # these assignments primarily for testing, diagnostics or reporting
@@ -866,16 +939,39 @@ module Asciidoctor
       end
 
       # NOTE document cannot control this behavior if safe >= SafeMode::SERVER
-      if !stream_output && doc.attr?('basebackend-html') && doc.attr?('copycss') &&
-          doc.attr?('linkcss') && DEFAULT_STYLESHEET_KEYS.include?(doc.attr('stylesheet'))
-        Helpers.require_library 'fileutils'
-        outdir = doc.attr('outdir')
-        stylesdir = doc.normalize_system_path(doc.attr('stylesdir'), outdir,
-            doc.safe >= SafeMode::SAFE ? outdir : nil)
-        Helpers.mkdir_p stylesdir
-        File.open(File.join(stylesdir, DEFAULT_STYLESHEET_NAME), 'w') {|f|
-          f.write Asciidoctor::HTML5.default_asciidoctor_stylesheet
-        }
+      if !stream_output && doc.safe < SafeMode::SECURE && (doc.attr? 'basebackend-html') &&
+          (doc.attr? 'linkcss') && (doc.attr? 'copycss')
+        copy_asciidoctor_stylesheet = DEFAULT_STYLESHEET_KEYS.include?(stylesheet = (doc.attr 'stylesheet'))
+        #copy_user_stylesheet = !copy_asciidoctor_stylesheet && (doc.attr? 'copycss')
+        copy_coderay_stylesheet = (doc.attr? 'source-highlighter', 'coderay') && (doc.attr 'coderay-css', 'class') == 'class'
+        copy_pygments_stylesheet = (doc.attr? 'source-highlighter', 'pygments') && (doc.attr 'pygments-css', 'class') == 'class'
+        if copy_asciidoctor_stylesheet || copy_coderay_stylesheet || copy_pygments_stylesheet
+          Helpers.require_library 'fileutils'
+          outdir = doc.attr('outdir')
+          stylesdir = doc.normalize_system_path(doc.attr('stylesdir'), outdir,
+              doc.safe >= SafeMode::SAFE ? outdir : nil)
+          Helpers.mkdir_p stylesdir if mkdirs
+          if copy_asciidoctor_stylesheet
+            File.open(File.join(stylesdir, DEFAULT_STYLESHEET_NAME), 'w') {|f|
+              f.write Asciidoctor::HTML5.default_asciidoctor_stylesheet
+            }
+          end
+
+          #if copy_user_stylesheet
+          #end
+
+          if copy_coderay_stylesheet
+            File.open(File.join(stylesdir, 'asciidoctor-coderay.css'), 'w') {|f|
+              f.write Asciidoctor::HTML5.default_coderay_stylesheet
+            }
+          end
+
+          if copy_pygments_stylesheet
+            File.open(File.join(stylesdir, 'asciidoctor-pygments.css'), 'w') {|f|
+              f.write Asciidoctor::HTML5.pygments_stylesheet(doc.attr 'pygments-style')
+            }
+          end
+        end
       end
       doc
     else
@@ -890,12 +986,11 @@ module Asciidoctor
   # options - a String, Array or Hash of options to control processing (default: {})
   #           String and Array values are converted into a Hash.
   #           See Asciidoctor::Document#initialize for details about options.
-  # block   - a callback block for handling include::[] directives
   #
   # returns the Document object if the rendered result String is written to a
   # file, otherwise the rendered result String
-  def self.render_file(filename, options = {}, &block)
-    Asciidoctor.render(File.new(filename), options, &block)
+  def self.render_file(filename, options = {})
+    Asciidoctor.render(File.new(filename), options)
   end
 
   # modules
@@ -915,7 +1010,7 @@ module Asciidoctor
   require 'asciidoctor/document'
   require 'asciidoctor/inline'
   require 'asciidoctor/lexer'
-  require 'asciidoctor/list_item'
+  require 'asciidoctor/list'
   require 'asciidoctor/path_resolver'
   require 'asciidoctor/reader'
   require 'asciidoctor/renderer'
